$:.unshift(File.dirname(__FILE__))

require 'des/feistel'
require 'des/key_schedule'

module DES
  IP_L = [58, 50, 42, 34, 26, 18, 10, 2, 60, 52, 44, 36, 28, 20, 12, 4, 62, 54, 46, 38, 30, 22, 14, 6, 64, 56, 48, 40, 32, 24, 16, 8]
  IP_R = [57, 49, 41, 33, 25, 17, 9, 1, 59, 51, 43, 35, 27, 19, 11, 3, 61, 53, 45, 37, 29, 21, 13, 5, 63, 55, 47, 39, 31, 23, 15, 7]
  FP = [40, 8, 48, 16, 56, 24, 64, 32, 39, 7, 47, 15, 55, 23, 63, 31, 38, 6, 46, 14, 54, 22, 62, 30, 37, 5, 45, 13, 53, 21, 61, 29,
        36, 4, 44, 12, 52, 20, 60, 28, 35, 3, 43, 11, 51, 19, 59, 27, 34, 2, 42, 10, 50, 18, 58, 26, 33, 1, 41, 9, 49, 17, 57, 25]

  class ::Array
    # @param [Array] another_ary
    # @return [Array]
    def xor(another_ary)
      self.map.with_index { |obj, i| (obj ^ another_ary[i]) }
    end
  end

  class Ctx
    attr_reader :data, :key

    def initialize(data, key)
      unless data.is_a?(DES::Block) and key.is_a?(DES::Block)
        raise 'DES::InvalidBlockFormat: Data and key must be a Block object.'
      end

      @data = data
      @key = key
    end

    def encrypt
      self.run(:encrypt)
    end

    def decrypt
      self.run(:decrypt)
    end

    protected

    # @param [Symbol] operation
    def run(operation)
      # l[0] is the IP_1_L permutation of the data block, l[1..16] are the results of each round of encryption
      l = []
      # r[0] is the IP_1_R permutation of the data block, r[1..16] are the results of each round of encryption
      r = []

      l << IP_L.collect { |p| data.bit_array[p - 1] }
      r << IP_R.collect { |p| data.bit_array[p - 1] }

      case operation
        when :encrypt
          k = KeySchedule.new(key.bit_array).sub_keys
        when :decrypt
          k = KeySchedule.new(key.bit_array).sub_keys.reverse
        else
          raise "Unknown operation - #{operation}"
      end

      16.times do |i|
        l << r[i]
        r << Feistel.run(r[i], k[i]).xor(l[i])
      end

      DES::Block.new(FP.collect { |p| (r.last + l.last)[p - 1] })
    end
  end

  class Block
    attr_reader :string, :bit_array

    alias_method :to_s, :string
    alias_method :to_a, :bit_array

    # @param [String, Array] input
    # @return [NilClass]
    def initialize(input)
      if input.is_a?(String)
        raise "DES::InvalidStringLength: Input String '#{input}' must contain (8) characters." unless input.length.eql?(8)

        @string = input
        @bit_array = input.unpack('B*').join.split('').collect { |b| b.to_i }
      elsif input.is_a?(Array)
        raise 'DES::InvalidArraySize: Input Array must contain (64) bits.' unless input.size.eql?(64)

        @string = input.join.lines.to_a.pack('B*')
        @bit_array = input
      else
        raise 'DES::InvalidFormat: Input must be a String or an Array.'
      end
    end
  end

  # @param [String, Array] data
  # @param [String, Array] key
  def self.encrypt(data, key)
    data_block = DES::Block.new(data)
    key_block = DES::Block.new(key)

    DES::Ctx.new(data_block, key_block).encrypt
  end

  # @param [DES::Block] data_block
  # @param [String] key
  def self.decrypt(data_block, key)
    key_block = DES::Block.new(key)
    DES::Ctx.new(data_block, key_block).decrypt
  end

end